/*
 *  Gtk-Parted - a front-end for libparted
 *  Copyright (C) 2000, 2001  Damien Genet
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * GUI functions
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <ctype.h>
//#include <sys/stat.h>
#include <errno.h>
#include <paths.h>

#include <parted/parted.h>
#include <gtk/gtk.h>

#include "../misc/misc.h"

#include "fstab.h"
#include "gped.h" 
#include "ctreeParted.h"

#define FSTYPE_EXT2		"ext2"
#define FSTYPE_FAT		"vfat"
#define FSTYPE_HFS		"hfs"
#define FSTYPE_LINUX_SWAP	"swap"
#define FSTYPE_NTFS		"ntfs"
#define FSTYPE_REISERFS		"reiserfs"
#define FSTYPE_HP_UFS		"ufs"
#define FSTYPE_UNKNOWN		"auto"

typedef struct _gpedFSTabList gpedFSTabList;

struct _gpedFSTabInfo
{
	gpedFSTabList	*fstab_list;
	gchar		*fstab_path;
	gboolean	 modified;
};

struct _gpedFSTabList
{
	gpedFSTabList	*next;
	gpedFSTabList	*prev;
	
	PedPartition	*part;
	gchar		*mount_point;
	gchar		*fstab_partname;
};


static gpedFSTabList*
_get_by_part                           (gpedFSTabList  *fstab_list,
                                        PedPartition   *part)
{
	gpedFSTabList	*walk;
	
	//g_return_val_if_fail (fstab_list, NULL);
	g_return_val_if_fail (part, NULL);
	
	for (walk = fstab_list; walk; walk = walk->next) {
		if (walk->part==part)
			break;
	}
	return walk;
}

static gpedFSTabList*
_get_by_fstab_partname                 (gpedFSTabList  *fstab_list,
                                        gchar          *part_name)
{
	gpedFSTabList	*walk;
	
	g_return_val_if_fail (fstab_list, NULL);
	g_return_val_if_fail (part_name, NULL);
	
	for (walk = fstab_list; walk; walk = walk->next) {
		if (str_are_equal (walk->fstab_partname, part_name))
			break;
	}
	return walk;
}

static gpedFSTabList*
_append                                (gpedFSTabList *fstab_list,
                                        gpedFSTabList *fsInfo)
{
	g_return_val_if_fail (fsInfo, NULL);
	
	fsInfo->prev = NULL;

	if (!fstab_list) {
		fsInfo->next = NULL;
	} else {
		fstab_list->prev = fsInfo;
		fsInfo->next = fstab_list;
	}
	
	return fsInfo;
}

static gpedFSTabList*
_gpedFSTabList_new                    (void)
{
	gpedFSTabList	*fsInfo;
	
	if ( !(fsInfo = g_malloc (sizeof (gpedFSTabList))) ) {
		g_printerr ("%s: allocationg gpedFSTabList.\n",
		            g_strerror (errno));
		return NULL;
	}

	return fsInfo;
}

static gpedFSTabList*
_append_new_entry                      (gpedFSTabList *fstab_list,
                                        PedPartition  *part,
                                        gchar         *mount_point,
                                        gchar         *fstab_partname)
{
	gpedFSTabList	*fsInfo;
	
	if ( !(fsInfo = _gpedFSTabList_new ()) )
		return NULL;
	
	fsInfo->part = part;
	fsInfo->mount_point = mount_point;
	fsInfo->fstab_partname = fstab_partname;
	
	return _append (fstab_list,  
	                fsInfo);
}	

gboolean
fstab_append_new_entry                 (gpedFSTabInfo *fstab_info,
                                        PedPartition  *part,
                                        gchar         *mount_point,
                                        gchar         *fstab_partname)
{
	gpedFSTabList	*fstab_list;

	g_return_val_if_fail (fstab_info, FALSE);
	
	if ( !(fstab_list = _append_new_entry (fstab_info->fstab_list,
	                                       part,
	                                       mount_point,
	                                       fstab_partname)) )
		return FALSE;
	
	fstab_info->fstab_list = fstab_list;
	return TRUE;
}

static void
_remove_entry                          (gpedFSTabList *fstab_list,
                                        PedPartition  *part)
{
	gpedFSTabList	*fsInfo;
	
	g_return_if_fail (fstab_list);
	g_return_if_fail (part);
	
	fsInfo = _get_by_part (fstab_list, part);
	g_return_if_fail (fsInfo);
	
	fsInfo->part = NULL;
}

void
fstab_remove_entry                     (gpedFSTabInfo *fstab_info,
                                        PedPartition  *part)
{
	g_return_if_fail (fstab_info);
	g_return_if_fail (part);
	
	if ( !fstab_info->fstab_list )
		return;
	
	_remove_entry (fstab_info->fstab_list,
	               part);
}

static void
_remove_all_disk_entries               (gpedFSTabList *fstab_list,
                                        PedDisk       *disk)
{
	PedPartition	*walk = NULL;
	
	g_return_if_fail (fstab_list);
	g_return_if_fail (disk);
	
	while ( (walk = ped_disk_next_partition (disk, walk)) )
		_remove_entry (fstab_list, walk);
}

void
fstab_remove_all_disk_entries          (gpedFSTabInfo *fstab_info,
                                        PedDisk       *disk)
{
	g_return_if_fail (fstab_info);
	g_return_if_fail (disk);
	
	if ( !fstab_info->fstab_list )
		return;
	
	_remove_all_disk_entries (fstab_info->fstab_list,
	                          disk);
}

static gchar*
_get_part_mount_point                  (gpedFSTabList *fstab_list,
                                        PedPartition  *part)
{
	gpedFSTabList	*fsInfo;
	
	g_return_val_if_fail (fstab_list, NULL);
	g_return_val_if_fail (part, NULL);
	
	if ( !(fsInfo = _get_by_part (fstab_list, part)) )
		return NULL;

	return fsInfo->mount_point;
}

gchar*
fstab_get_part_mount_point             (gpedFSTabInfo *fstab_info,
                                        PedPartition  *part)
{
	g_return_val_if_fail (fstab_info, NULL);
	g_return_val_if_fail (part, NULL);

	if ( !fstab_info->fstab_list )
		return NULL;
	
	return _get_part_mount_point (fstab_info->fstab_list,
	                              part);
}

static gboolean
_set_part_mount_point                  (gpedFSTabList *fstab_list,
                                        PedPartition  *part,
                                        gchar         *mount_point)
{
	gpedFSTabList	*fsInfo;

	g_return_val_if_fail (fstab_list, FALSE);
	g_return_val_if_fail (part, FALSE);
	
	if ( !(fsInfo = _get_by_part (fstab_list, part)) )
		return FALSE;
	
	if (fsInfo->mount_point)
		g_free (fsInfo->mount_point);
	fsInfo->mount_point = mount_point;

	return TRUE;
}

gboolean
fstab_set_part_mount_point             (gpedFSTabInfo *fstab_info,
                                        PedPartition  *part,
                                        gchar         *mount_point)
{
	g_return_val_if_fail (fstab_info, FALSE);
	g_return_val_if_fail (part, FALSE);
	
	return _set_part_mount_point (fstab_info->fstab_list,
	                              part,
	                              mount_point);
}

static void
_get_partname_mountpoint               (gchar          *line,
                                        gchar         **part_name,
                                        gchar         **mount_point)
{
	g_return_if_fail (line);
	g_return_if_fail (part_name);
	g_return_if_fail (mount_point);
	
	*part_name = str_dup_get_nth_word (line, 1);
	*mount_point = str_dup_get_nth_word (line, 2);
}

static gchar*
_get_disk_minor                        (gchar          *string,
                                        guint          *minor)
{
	gchar	*walk;
	gchar	*string_end = string + strlen (string) - 1;
	
	g_return_val_if_fail (string, NULL);
	g_return_val_if_fail (minor, NULL);
	
	walk = string_end;
	while (walk > string && isdigit (*walk))
		walk--;
	
	if (walk == string) {
		return NULL;
	} else if (walk < string_end) {
		*minor = (gint) strtoul (walk+1, NULL, 10);
	} else {
		*minor = 0;
	}
		
	return g_strndup (string, walk - string + 1);
}

static gpedFSTabList*
_read_line                             (gchar          *line,
                                        gpedDeviceInfo *dev_list,
                                        gpedFSTabList  *fstab_list)
{
	PedPartition	*part;
	gchar		*part_name;
	gchar		*dev_name = NULL;
	gchar		*mount_point;
	guint		 part_minor = 0;
	
	g_return_val_if_fail (line, fstab_list); 
	
	if (str_is_shell_comment_or_empty (line))
		return fstab_list;
	
	printf ("\n%s", line);
	_get_partname_mountpoint (line, &part_name, &mount_point);
	
	printf ("part_name: %s\n", part_name);
	dev_name = _get_disk_minor (part_name, &part_minor);
	part = gped_get_part (dev_name, part_minor, dev_list);
	printf ("disk: %s, minor: %u\n", dev_name, part_minor);
	g_free (dev_name);
	
	if (!part) {
		printf ("no corresponding PedPartition found, skip entry\n");
		return fstab_list;
	}
	
	printf ("mount_point: %s\n", mount_point);
	
	return _append_new_entry (fstab_list,
	                          part,
	                          mount_point,
	                          part_name);
}

static gpedFSTabList*
_build_list                            (gchar          *fstab_path,
                                        gpedDeviceInfo *dev_list)
{
	gpedFSTabList	*fstab_list = NULL;
	FILE		*f;
	gchar		 buffer[BUFFER_SIZE];
	
	g_return_val_if_fail (fstab_path, NULL);
	g_return_val_if_fail (dev_list, NULL);
	
	if ( !(f = fopen (fstab_path, "r")) ) {
		g_printerr ("%s: opening %s.\n",
		            g_strerror (errno),
		            fstab_path);
		return NULL;
	}

	while ( fgets (buffer, BUFFER_SIZE, f) )
		fstab_list = _read_line (buffer, dev_list, fstab_list);
	
	if ( fclose (f) )
		g_printerr ("%s: closing %s.\n",
		            g_strerror (errno),
		            fstab_path);
	
	return fstab_list;
}

static gpedFSTabInfo*
_gpedFSTabInfo_new                    (void)
{
	gpedFSTabInfo	*fstab_info;
	
	if ( !(fstab_info = g_malloc (sizeof (gpedFSTabInfo))) ) {
		g_printerr ("%s: allocating gpedFSTabInfo.\n",
		            g_strerror (errno));
		return NULL;
	}

	return fstab_info;
}

static gchar*
_get_fstab_path                        (gchar         *fstab_path)
{
	if ( !fstab_path )
		return _PATH_MNTTAB;

	return fstab_path;
}

gpedFSTabInfo*
fstab_build_info                       (gchar          *fstab_path,
                                        gpedDeviceInfo *dev_list)
{
	gpedFSTabInfo	*fstab_info;

	g_return_val_if_fail (dev_list, NULL);
	
	if ( !(fstab_info = _gpedFSTabInfo_new ()) ) 
		return NULL;
	
	fstab_info->fstab_path = _get_fstab_path (fstab_path);
	printf ("    fstab_path: %s\n", fstab_info->fstab_path);
	
	printf ("    build fstab list\n");
	fstab_info->fstab_list = _build_list (fstab_info->fstab_path, 
	                                      dev_list);

	fstab_info->modified = FALSE;

	return fstab_info;
}

void
fstab_change_done                      (gpedInfo       *info)
{
	g_return_if_fail (info);
	g_return_if_fail (info->ctreeParted);
	g_return_if_fail (info->fstab_info);
	
	info->fstab_info->modified = TRUE;
	ctreeParted_refresh (GTK_CTREE (info->ctreeParted),
	                     info);
}

gboolean
fstab_get_changes                      (gpedFSTabInfo  *fstab_info)
{
	g_return_val_if_fail (fstab_info, FALSE);
	
	return fstab_info->modified;
}

static void
_save_line                             (FILE           *f_write,
                                        gchar          *line,
                                        gpedFSTabList  *fstab_list)
{
	gpedFSTabList	*fsInfo;
	gchar		*part_name;
	gchar		*mount_point;
	gchar		*dev_name = NULL;
	guint		 part_minor = 0;
	guint            length;

	g_return_if_fail (f_write);
	g_return_if_fail (line); 

	if (str_is_shell_comment_or_empty (line)) {
		fputs (line, f_write);
		return;
	}
	
	_get_partname_mountpoint (line, &part_name, &mount_point);
	dev_name = _get_disk_minor (part_name, &part_minor);
	if ( !(fsInfo
	        = _get_by_fstab_partname (fstab_list,
	                                  part_name)) ) {
		/* not present in fstab_list, keep */
		fputs (line, f_write);
	} else if (!fsInfo->part) {
		/* part deleted */
	} else if ( str_are_equal (fsInfo->mount_point, mount_point)
	            && !strcmp (fsInfo->part->geom.disk->dev->path, dev_name)
	            && fsInfo->part->num == part_minor)  {
		/* unmodified fstab infos */
		fputs (line, f_write);
	} else {
		fprintf (f_write, "%s%i\t%s\t%s",
		         fsInfo->part->geom.disk->dev->path,
		         fsInfo->part->num,
		         fsInfo->mount_point, 
		         str_get_nth_word (line, 3, &length));
	}
	g_free (dev_name);
	g_free (part_name);
	g_free (mount_point);
}

static gchar*
_get_fstab_fs_name                     (const gchar    *parted_fs_name)
{
	g_return_val_if_fail (parted_fs_name, NULL);

	if (!strcmp (parted_fs_name, "ext2"))
		return FSTYPE_EXT2;
	else if (!strcmp (parted_fs_name, "FAT"))
		return FSTYPE_FAT;
	else if (!strcmp (parted_fs_name, "hfs"))
		return FSTYPE_HFS;
	else if (!strcmp (parted_fs_name, "linux-swap"))
		return FSTYPE_LINUX_SWAP;
	else if (!strcmp (parted_fs_name, "ntfs"))
		return FSTYPE_NTFS;
	else if (!strcmp (parted_fs_name, "reseirfs"))
		return FSTYPE_REISERFS;
	else if (!strcmp (parted_fs_name, "hp-ufs"))
		return FSTYPE_HP_UFS;
	else
		return FSTYPE_UNKNOWN;
}

static void
_save_new                              (FILE           *f_write,
                                        gpedFSTabList  *fstab_list)
{
	gpedFSTabList	*walk;
	
	g_return_if_fail (f_write);
	
	for (walk = fstab_list; walk; walk = walk->next) {
		if (!walk->fstab_partname) {
			/* new entry */
			fprintf (f_write, "%s%i\t%s\t%s\tdefaults\t0\t0\n",
			         walk->part->geom.disk->dev->path,
			         walk->part->num,
			         walk->mount_point,
			         _get_fstab_fs_name (walk->part->fs_type->name));
		}
	}
}

gboolean
fstab_save                             (gpedFSTabInfo *fstab_info)
{
	FILE	*f_read, *f_write;
	gchar	 buffer[BUFFER_SIZE];
	gchar	*fstab_save;
	gchar	*fstab_new;

	g_return_val_if_fail (fstab_info, FALSE);
	g_return_val_if_fail (fstab_info->fstab_path, FALSE);
	
	if ( !fstab_get_changes (fstab_info) ) {
		printf ("%s is not modified\n", fstab_info->fstab_path);
		return TRUE;
	}
	
	if ( !(f_read = fopen (fstab_info->fstab_path, "r")) ) {
		g_printerr ("%s: opening %s for read.\n",
		            g_strerror (errno),
		            fstab_info->fstab_path);
		goto error;
	}
	fstab_new  = g_strconcat (fstab_info->fstab_path, ".gtkparted", NULL);
	if ( !(f_write = fopen (fstab_new, "w+")) ) {
		g_printerr ("%s: opening %s for write.\n",
		            g_strerror (errno),
		            fstab_new);
		goto error_free_fstab_new;
	}
	
	while ( fgets (buffer, BUFFER_SIZE, f_read) )
		_save_line (f_write, buffer, fstab_info->fstab_list);
	_save_new (f_write, fstab_info->fstab_list);
	
	if ( fclose (f_read) )
		g_printerr ("%s: closing %s.\n",
		            g_strerror (errno),
		            fstab_info->fstab_path);
	if ( fclose (f_write) )
		g_printerr ("%s: closing %s.\n",
		            g_strerror (errno),
		            fstab_new);

	fstab_save  = g_strconcat (fstab_info->fstab_path, "~", NULL);
	printf ("saving fstab to %s\n", fstab_save);
	if ( rename (fstab_info->fstab_path, fstab_save) ) {
		g_printerr ("%s: moving fstab %s to %s.\n",
		            g_strerror (errno),
		            fstab_info->fstab_path,
		            fstab_save);
		goto error_free_fstab_save;
	}
	g_free (fstab_save);
	
	if ( rename (fstab_new, fstab_info->fstab_path) ) {
		g_printerr ("%s: moving new fstab %s to %s.\n",
		            g_strerror (errno),
		            fstab_new,
		            fstab_info->fstab_path);
		goto error_free_fstab_new;
	}
	g_free (fstab_save);

	fstab_info->modified = FALSE;
	
	return TRUE;

error_free_fstab_save:
	g_free (fstab_save);
error_free_fstab_new:
	g_free (fstab_new);
error:
	return FALSE;
}
